%{
/* csv.lex - (f)lex lexer for comma separated values (CSV) files
**
** Copyright (C) 2022, 2023  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "y.tab.h"

extern union YYSTYPE yylval;

int yywrap(void);

unsigned int	csvLineNumber = 1;

static char *fixQstring(char *source, size_t length);

%}
EOL	(\n|\r\n?)
DQ	(\")
%x QSTRING QEND
%%
,			return COMMA;

{EOL}			{
			  ++csvLineNumber;
			  return EOL;
			}

{DQ}			{ /* standalone doublequote starts a quoted string */
			  BEGIN QSTRING;
			}

[^,"[:cntrl:]]+		{ /* valid characters in an unquoted string */
			  yylval.String = strdup(yytext);
			  return STRING;
			}

<QSTRING>{EOL}		yymore();	/* allow line ending within quoted string */

<QSTRING>[^"[:cntrl:]]+	yymore();

<QSTRING>{DQ}{DQ}	yymore();

<QSTRING>{DQ}		{ /* unescaped doublequote terminates the quoted string */
			  yyless(yyleng-1);	/* requeue the terminating dquote */
			  BEGIN QEND;		/* let QEND process it */

			  /* now, fix up the various escapes and return a STRING */
			  yylval.String = fixQstring(yytext,yyleng);
			  return STRING;
			}

<QSTRING>.		{ /* unexpected or illegal character in quoted string */
			  yyless(yyleng-1);	/* requeue matched character */
			  BEGIN INITIAL;	/* discards corrupt quoted string */ 
			}

<QEND>{DQ}		{ BEGIN INITIAL; /* discard doublequote */}

.			return *yytext;

%%

int yywrap(void) { return 1; }

/*
** fixQstring()
**
** This function isolates the csv quoted string extracted 
** at the <QSTRING>{DQ} rule, and transforms certain single
** and multibyte segments into single characters
** Currently recognized:
**   input ""	(two double quotes) becomes " (one doublequote)
**   input \r\n (MSDOS line ending) becomes \n (C line ending)
**   input \r   (MAC line ending) becomes \n (C line ending)
**
** The function returns a pointer to the isolated and transformed
** string (which must be free()d by the recipient of the string
*/
static char *fixQstring(char *source, size_t length)
{
  char	*tmp = strndup(source,length),
	*src,
	*dst;

  for (src = dst = tmp; src && *src; ++src)
  {
    if ((strncmp((const char *)src,"\"\"",2) == 0)	/* escaped doublequote */
    ||  (strncmp((const char *)src,"\r\n",2) == 0))	/* MSDOS text line ending */
      ++src;						/*   2nd character is always the one we want */
    else if (*src == '\r')				/* old MAC line ending */
      *src = '\n';					/*   substitute \n for \r */

    *dst++ = *src;					/* copy from source to dest */
  }
  if (dst) *dst = 0;					/* string terminate dest */

  return tmp;
}

/*
** $Revision: 1.15 $
**
** $Log: csv.lex,v $
** Revision 1.15  2024/11/05 19:13:53  lpitcher
** rename the Q2 expression to DQ
** reduce recognized line endings to \r\n, \r, and \n
** move doublequoted-string isolation logic to own function
** change doublequoted-string isolation logic to properly
** transform double-doublequotes, and various line-endings
**
** Revision 1.14  2024/11/04 13:11:38  lpitcher
** "define" some common regular expressions
**
** Revision 1.13  2024/11/03 23:09:00  lpitcher
** allow for alternate line endings (cr or lf or crlf or lfcr) in quoted strings
**
** Revision 1.12  2024/11/02 18:39:52  lpitcher
** added RCS revision number and RCS log
**
**
*/