%{
/* csv.y - yacc grammar for comma separated values (CSV) files
**
** Copyright (C) 2022, 2023  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
*/
  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>

  #include "csvtool.h"

  int yyerror(char *string);

  extern int yylex(void);
  extern int yywrap(void);
  extern char *yytext;
  extern unsigned int csvLineNumber;

  struct csvCell {
    struct csvCell	*cellLink;
    char		*cellString;
  };

  struct csvRow {
    struct csvRow	*rowLink;
    size_t		cellCount;
    struct csvCell	*cellList;
  };

  struct csvTbl {
    size_t		rowCount;
    struct csvRow	*rowList;
  };

  static struct csvTbl	*csvTable;

  static struct csvCell	*csvNewCell(struct csvCell *link, char *value);
  static struct csvRow	*csvNewRow(struct csvRow *link, struct csvCell *list);
  static struct csvTbl	*csvNewTbl(struct csvRow *list);
%}
%union {
  char			*String;
  struct csvCell	*Cell;
  struct csvRow		*Row;
  struct csvTbl		*Tbl;
}
%token	EOL COMMA
%token	<String> STRING
%type	<String> value
%type	<Cell>	values line
%type	<Row>	lines
%type	<Tbl>	csv
%%
csvfile	:csv			{ csvTable = $1; }
	;

csv	: lines values		{
				  struct csvRow *rows;

				  if (!(rows = csvNewRow($1,$2)))	YYABORT;
				  if (!($$ = csvNewTbl(rows)))		YYABORT; 
				}
	| lines			{ if (!($$ = csvNewTbl($1)))		YYABORT; }
	| error			{ $$ = NULL; YYABORT; }
	;

lines	: lines line		{ if (!($$ = csvNewRow($1,$2)))		YYABORT; }
	| line			{ if (!($$ = csvNewRow(NULL,$1)))	YYABORT; }
	;

line	: values EOL		{ $$ = $1; }
	;

values	: values COMMA value	{ if (!($$ = csvNewCell($1,$3)))	YYABORT; }
	| value			{ if (!($$ = csvNewCell(NULL,$1)))	YYABORT; }
	;

value	:			{ $$ = NULL; }
	| STRING		{ if (!($$ = $1)) { yyerror("Out of memory"); YYABORT; } }
	;

%%
/*
******************************* FUNCTIONS *************************************
*/

int yyerror(char *string)
{
  fprintf(stderr,"csvOpen(): %s while processing CSV row %u\n",string,csvLineNumber);
  return 0;
}

static struct csvCell	*csvNewCell(struct csvCell *link, char *value)
{
  struct csvCell *this;

  if (this = malloc(sizeof *this))
  {
    this->cellLink = link;
    this->cellString = value;
  }
  else yyerror("Out of memory");
  return this;
}

static struct csvRow	*csvNewRow(struct csvRow *link, struct csvCell *list)
{
  struct csvRow *this;

  if (this = malloc(sizeof *this))
  {
    size_t		count = 0;
    struct csvCell	*node,
			*prev = NULL;

    this->rowLink = link;

    /*
    ** NB: We had stored the cell list in LIFO order.
    **     Now, we will reorder it into FIFO order
    **     to make later data retrievals easier.
    **     This also allows us to count the # of cells
    **     for row reporting purposes.
    */
    for (node = list, count = 0; node;)
    {
      struct csvCell	*temp;
      temp = node->cellLink;
      node->cellLink = prev;
      ++count;
      prev = node;
      node = temp;
    }
    this->cellCount = count;
    this->cellList = prev;
  }
  else yyerror("Out of memory");
  return this;
}

static struct csvTbl	*csvNewTbl(struct csvRow *list)
{
  struct csvTbl	*this;

  if (this = malloc(sizeof *this))
  {
    size_t		count = 0;
    struct csvRow	*node,
			*prev = NULL;

    /*
    ** NB: We had stored the row list in LIFO order.
    **     Now, we will reorder it into FIFO order
    **     to make later data retrievals easier.
    **     This also allows us to count the # of rows
    **     for table reporting purposes
    */
    for (node = list, count = 0; node;)
    {
      struct csvRow	*temp;

      temp = node->rowLink;
      node->rowLink = prev;
      count++;
      prev = node;
      node = temp;
    }
    this->rowCount = count;
    this->rowList  = prev;
  }
  else yyerror("Out of memory");
  return this;
}

/*
** PUBLIC API
*/

size_t	csvRowCount(void *handle)
{
  return ((struct csvTbl *)handle)->rowCount;
}

size_t	csvCellCount(void *handle, size_t row)
{
  struct csvRow *tmp;
  size_t count;

  for (tmp = ((struct csvTbl *)handle)->rowList, count = 0;
       tmp && (count < row);
       tmp = tmp->rowLink, count++) continue;
  if (tmp)
    count = tmp->cellCount;
  else
    count = 0;
  return count;
}

char	*csvCellData(void *handle, size_t row, size_t cell)
{
  struct csvRow *tmp;
  size_t	count;
  char 		*value = NULL;

  for (tmp = ((struct csvTbl *)handle)->rowList, count = 0;
       tmp && (count < row);
       tmp = tmp->rowLink, count++) continue;
  if (tmp)
  {
    struct csvCell *node;
    for (node = tmp->cellList, count = 0;
	 node && (count < cell);
	 node = node->cellLink, count++) continue;
    if (node) value = node->cellString;
  }

  return value;
}

void	csvClose(void *handle)
{
  struct csvRow	*row;

  for (row = ((struct csvTbl *)handle)->rowList; row;)
  {
    struct csvRow	*discard;
    struct csvCell	*cell;

    for (cell = row->cellList; cell; )
    {
      struct csvCell	*discard;

      free(cell->cellString);
      discard = cell;
      cell = cell->cellLink;
      free(discard);
    }

    discard = row;
    row = row->rowLink;
    free(discard);
  }

  free(handle);
}

void *csvOpen(char *fname)
{
  extern FILE *yyin;
  struct csvTbl	*handle;

  if (yyin = fname ? fopen(fname,"rb") : fdopen(fileno(stdin),"rb"))
  {
    csvLineNumber = 1;
    handle = yyparse() ? NULL: csvTable;
    fclose(yyin);
  }

  return (void *)handle;
}

/*
** $Revision: 1.14 $
**
** $Log: csv.y,v $
** Revision 1.14  2024/11/05 19:35:05  lpitcher
** refactored csvOpen() function
**
** Revision 1.13  2024/11/05 19:13:53  lpitcher
** open input csv as "rb" to permit proper isolation
** of msdos, mac line endings
**
** Revision 1.12  2024/11/02 18:39:52  lpitcher
** added RCS revision number and RCS log
**
**
*/
